// Copyright Vygintas D., 2003. All rights reserved.
//
// Permission to use, copy, modify, and distribute this software for any
// purpose is hereby granted without fee, provided that this copyright and
// permissions notice appear in all copies and derivatives.
//
// This software is provided "as is" without express or implied warranty.

#ifndef _DETAIL_INLINE_HPP_
#define _DETAIL_INLINE_HPP_

#include <cstddef>

// FIXES for broken compilers
#include <boost/config.hpp>
#include <boost/static_assert.hpp>

#include <boost/utility.hpp>
#include <boost/call_traits.hpp>

#if defined(BOOST_NO_TEMPLATE_PARTIAL_SPECIALIZATION)
#   include <boost/mpl/if.hpp>
#endif

namespace math {

    namespace detail {
    
        // One may wonder why template parameter T is suplied
        // when applicators deal with Input/Output iterators.
        // Rationale behind this desicion is simple and clear -
        // std::iterator_traits template does not work properly
        // with pointer iterators on compilers that do not
        // support partial template specializations. 
        // Boost iterator traits is not much helping either.
        // So, trying to avoid unnecessary mess I just introduced
        // template parameter for iterator value type.

        template< typename T1, typename T2, typename T3 >
        struct applicator
        {
            typedef T1      value_type;
            typedef T2      first_argument_type;
            typedef T3      second_argument_type;

            typedef typename boost::call_traits<T2>::param_type 
                fast_first_argument_type;
            typedef typename boost::call_traits<T3>::param_type 
                fast_second_argument_type;

            typedef void    result_type;
        };

        template< typename T, typename InputIter, typename OutputIter >
        struct assign : applicator<T,InputIter,OutputIter>
        {
            inline static void apply( InputIter iit, OutputIter oit )
                // needs to be clarified,
                //  perhaps static_cast<T>(*iit) 
                //  would be more appriopriate solution
                { *oit = T(*iit); }
        };

        template< typename T, typename OutputIter >
        struct assign_v : applicator<T,OutputIter,T>
        {
            typedef typename applicator<T,OutputIter,T>::
                fast_second_argument_type fast_arg2_t;
                            
            inline static void apply( OutputIter oit, fast_arg2_t v )
                { *oit = v; }
        };

        template< typename T, typename InputIter >
        struct and_comp : applicator<T,InputIter,InputIter>
        {
            typedef bool result_type;
            inline static bool comp( bool a, bool b )
                { return (a && b); }
        };

        template< typename T, typename InputIter >
        struct equal : and_comp<T, InputIter>
        {
            inline static bool apply( InputIter a, InputIter b )
                { return (*a == *b); }
        };

        template< typename T, typename InputIter >
        struct less : and_comp<T, InputIter>
        {
            inline static bool apply( InputIter a, InputIter b )
                { return (*a < *b); }
        };

        // ------------------------------------------------------------------
        // N times (>=1) applies an applicator to the sequence of known size.
        // ------------------------------------------------------------------
        template< typename Predicate, std::size_t N >
        struct inline_;

        namespace aux
        {
            // just aux/helper structs
            template< typename Predicate >
            struct last_op_
            {
                typedef typename Predicate::first_argument_type arg1_t;
                typedef typename Predicate::second_argument_type arg2_t;
                typedef typename Predicate::fast_second_argument_type fast_arg2_t;
                typedef typename Predicate::result_type res_t;

                inline static void apply_seq( arg1_t iit, arg2_t oit )
                { Predicate::apply(iit, oit); }

                inline static void apply_value( arg1_t oit, fast_arg2_t v )
                { Predicate::apply(oit, v); }

                inline static res_t compute_seq( arg1_t iit1, arg2_t iit2 )
                { return Predicate::apply(iit1, iit2); }
            };
        
            template< typename Predicate, std::size_t N >
            struct inline_ops_
            {
                typedef inline_<Predicate, N - 1> nextf;

                typedef typename Predicate::first_argument_type arg1_t;
                typedef typename Predicate::second_argument_type arg2_t;
                typedef typename Predicate::fast_second_argument_type fast_arg2_t;
                typedef typename Predicate::result_type res_t;

                inline static void apply_seq( arg1_t iit, arg2_t oit )
                {
                    Predicate::apply(iit, oit);
                        nextf::apply_seq( ++iit, ++oit );
                }

                inline static void apply_value( arg1_t oit, fast_arg2_t v )
                {
                    Predicate::apply(oit, v);
                        nextf::apply_value( ++oit, v );
                }

                inline static res_t compute_seq( arg1_t iit1, arg2_t iit2 )
                {
                    return( Predicate::comp( Predicate::apply(iit1, iit2),
                        // MSVC uses changed iterator values for
                        // applicator, so this solution seems to be
                        // quite harmless for performance and code safety
                        nextf::compute_seq(boost::next(iit1), boost::next(iit2)) ) );
                }
            };
        } // namespace aux

#if !defined(BOOST_NO_TEMPLATE_PARTIAL_SPECIALIZATION)
        template< typename Predicate, std::size_t N >
        struct inline_ : aux::inline_ops_< Predicate, N > {};

        template< typename Predicate >
        struct inline_< Predicate, 1 > : aux::last_op_<Predicate> {};

        template< typename Predicate >
        struct inline_< Predicate, 0 > {}; // prevent from nonsensical instantiations
#else

        template< typename Predicate, std::size_t N >
        struct inline_ :
            boost::mpl::if_c<
                (N > 1)
                , aux::inline_ops_<Predicate,N>
                , aux::last_op_<Predicate> >::type
        {
            // prevent from nonsensical instantiations
            BOOST_STATIC_ASSERT( N > 0 );
        };

#endif
    } // math::detail

} // namespace math

#endif // _DETAIL_INLINE_HPP_
