// Copyright Vygintas D., 2003. All rights reserved.
//
// Permission to use, copy, modify, and distribute this software for any
// purpose is hereby granted without fee, provided that this copyright and
// permissions notice appear in all copies and derivatives.
//
// This software is provided "as is" without express or implied warranty.

#ifndef _DETAIL_TINY_VECTOR_HPP_
#define _DETAIL_TINY_VECTOR_HPP_

#include "inline_.hpp"

namespace math {

    namespace detail {
    
        template< typename T, typename OutputIter >
        struct mul_v : applicator<T,OutputIter,T>
        {
            typedef typename applicator<T,OutputIter,T>::
                fast_second_argument_type fast_arg2_t;

            inline static void apply( OutputIter oit, fast_arg2_t v )
                { *oit *= v; }
        };
        
        template< typename T, typename OutputIter >
        struct div_v : applicator<T,OutputIter,T>
        {
            typedef typename applicator<T,OutputIter,T>::
                fast_second_argument_type fast_arg2_t;

            inline static T divisor( fast_arg2_t v )
                { return v; }
            inline static void apply( OutputIter oit, fast_arg2_t v )
                { *oit /= v; }
        };

        template< typename T, typename InputIter, typename OutputIter >
        struct negate_assign : applicator<T,InputIter,OutputIter>
        {
            inline static void apply( InputIter iit, OutputIter oit )
                { *oit = -(*iit); }
        };

        template< typename T, typename InputIter, typename OutputIter >
        struct add : applicator<T,InputIter,OutputIter>
        {
            inline static void apply( InputIter iit, OutputIter oit )
                { *oit += *iit; }
        };

        template< typename T, typename InputIter, typename OutputIter >
        struct sub : applicator<T,InputIter,OutputIter>
        {
            inline static void apply( InputIter iit, OutputIter oit )
                { *oit -= *iit; }
        };

        template< typename T, typename InputIter >
        struct dot_pr : applicator<T,InputIter,InputIter>
        {
            typedef T result_type;
            typedef typename boost::call_traits<T>::param_type fast_arg_t;
            
            inline static T apply( InputIter a, InputIter b )
                { return ((*a) * (*b)); }

            inline static T comp( fast_arg_t a, fast_arg_t b )
                { return (a + b); }
        };
        

        // division optimization for float/double types
        template<>
        struct div_v<float, float*>
            : applicator<float,float*,float>
        {
            inline static float divisor( const float v )
                { return 1.f / v; }
            inline static void apply( float* oit, const float v )
                { *oit *= v; }
        };

        template<>
        struct div_v<double, double*>
            : applicator<double,double*,double>
        {
            inline static double divisor( const double v )
                { return 1.0 / v; }
            inline static void apply( double* oit, const double v )
                { *oit *= v; }
        };


#ifdef BOOST_MSVC
#   pragma warning( disable : 4100 ) // unreferenced formal parameter
#endif

        // vector downgrade / upgrade
#if !defined(BOOST_NO_TEMPLATE_PARTIAL_SPECIALIZATION)
        namespace aux
        {
            template< class T, class T2, bool Flag >
            struct downgrade_upgrade;

            template< class T, class T2 >
            struct downgrade_upgrade< T, T2, true >
            {
                // upgrading rhs vector
                inline static T& apply( T& v, const T2& rhs )
                    { v[T2::static_size] = 1; return v; }
            };

            template< class T, class T2 >
            struct downgrade_upgrade< T, T2, false >
            {
                // downgrading rhs vector
                inline static T& apply( T& v, const T2& rhs )
                    { return ( v /= rhs[T::static_size] ); }
            };
        }

        template< class T, class T2 >
        struct change_grade :
            aux::downgrade_upgrade< T, T2, (T::static_size > T2::static_size) >
        {};
#else
        namespace aux
        {
            template< class T, class T2 >
            struct upgrade_rhs
            {
                // upgrading rhs vector
                inline static T& apply( T& v, const T2& rhs )
                    { v[T2::static_size] = 1; return v; }
            };

            template< class T, class T2 >
            struct degrade_rhs
            {
                // downgrading rhs vector
                inline static T& apply( T& v, const T2& rhs )
                    { return ( v /= rhs[T::static_size] ); }
            };
        }

        template< class T, class T2 >
        struct change_grade
        {
            // Sorry, MSVC .NET does a bad job to
            // compare enumerators when no variables
            // of abstract types are declared, c'est la vie
            inline static T& apply( T& v, const T2& rhs )
            { 
                typedef typename boost::mpl::if_c<
                    (T::static_size > T2::static_size)
                    , aux::upgrade_rhs<T, T2>
                    , aux::degrade_rhs<T, T2> >::type Pred;
                return Pred::apply( v, rhs ); 
            }
        };
#endif

#ifdef BOOST_MSVC
#   pragma warning( default : 4100 ) // unreferenced formal parameter
#endif

    } // math::detail

} // namespace math

#endif // _DETAIL_TINY_VECTOR_HPP_
