// Copyright Vygintas D., 2003. All rights reserved.
//
// Permission to use, copy, modify, and distribute this software for any
// purpose is hereby granted without fee, provided that this copyright and
// permissions notice appear in all copies and derivatives.
//
// This software is provided "as is" without express or implied warranty.

#ifndef _TINY_VECTOR_HPP_
#define _TINY_VECTOR_HPP_

// In order to change the maximum number of params that tiny_vector c-tor
// supports, you just simply change the MAX_TINY_VECTOR_CTOR_PARAMS
// definition and recompile
#if !defined(MAX_TINY_VECTOR_CTOR_PARAMS)
#   define MAX_TINY_VECTOR_CTOR_PARAMS 11
#endif

#include <stdexcept>

#include <boost/preprocessor/arithmetic/inc.hpp>
#include <boost/preprocessor/punctuation/comma_if.hpp>
#include <boost/preprocessor/repetition.hpp>

#include <boost/operators.hpp>

#include "sqrt.hpp"
#include "detail/tiny_vector.hpp"

namespace math {

    template<class T, std::size_t N>
    class tiny_vector    // note: private inheritance is OK here!
        : boost::addable< tiny_vector<T,N>          // vector + vector
        , boost::subtractable< tiny_vector<T,N>     // vector - vector
        , boost::dividable2< tiny_vector<T,N>, T    // vector / T
        , boost::multipliable2< tiny_vector<T,N>, T // vector * T, T * vector
        , boost::xorable< tiny_vector<T,N>
        , boost::equality_comparable< tiny_vector<T,N>
        , boost::less_than_comparable< tiny_vector<T,N>
        > > > > > > >
    {
    protected:
        T data_[N];    // fixed-size array of elements of type T

        typedef tiny_vector<T,N>            self_t;
        typedef typename boost::call_traits<T>::param_type fast_arg_t;

    public:
        // type definitions
        typedef T                           value_type;
        typedef T*                          iterator;
        typedef const T*                    const_iterator;
        typedef T&                          reference;
        typedef const T&                    const_reference;
        typedef std::size_t                 size_type;
        typedef std::ptrdiff_t              difference_type;

        typedef typename sqrt_t<T>::type    length_type;

        // The elements of the vector are left uninitialized.
        inline tiny_vector() {}

        // The elements of vector v are copied.
        inline tiny_vector( const self_t& v ) {
            detail::inline_<
                detail::assign<T, const_iterator, iterator>, N>
            ::apply_seq(v.begin(), begin());
        }

        // All elements are initialized to value.
        explicit inline tiny_vector( fast_arg_t v ) {
            detail::inline_<
                detail::assign_v<T, iterator>, N>
            ::apply_value(begin(), v);
        }

        // The vector is initialized with the list of values given.
        // These constructors are provided for up to N=11.
        // tiny_vector( fast_arg_t value1, fast_arg_t value2, ...)

#define TINY_VECTOR_CTOR_AUX(Z, N, _) data_[N] = v##N;
#define TINY_VECTOR_CTOR(Z, N, VSIZE) \
        inline tiny_vector(fast_arg_t v0 BOOST_PP_COMMA_IF(N) BOOST_PP_ENUM_SHIFTED_PARAMS(N, fast_arg_t v) ) \
        {                                                    \
            BOOST_STATIC_ASSERT( VSIZE == N );               \
            BOOST_PP_REPEAT( N, TINY_VECTOR_CTOR_AUX, _ )    \
        }                                                    \
        /**/

        // creates c-tors from 2 params, to MAX_TINY_VECTOR_CTOR_PARAMS params;
        // it goes without saying that MAX_TINY_VECTOR_CTOR_PARAMS
        // cannot be < 2
        BOOST_PP_REPEAT_FROM_TO(
            2, BOOST_PP_INC(MAX_TINY_VECTOR_CTOR_PARAMS),
        TINY_VECTOR_CTOR, N )

#undef TINY_VECTOR_CTOR
#undef TINY_VECTOR_CTOR_AUX

        // iterator support
        inline iterator begin() { return data_; }
        inline const_iterator begin() const { return data_; }
        inline iterator end() { return data_+N; }
        inline const_iterator end() const { return data_+N; }

        // operator[]
        inline reference operator[](size_type i) { return data_[i]; }
        inline const_reference operator[](size_type i) const { return data_[i]; }

        // at() with range check
        reference at(size_type i) { rangecheck(i); return data_[i]; }
        const_reference at(size_type i) const { rangecheck(i); return data_[i]; }

        // size is constant
        BOOST_STATIC_CONSTANT( std::size_t, static_size = N );
        inline static size_type size() { return N; }
        inline static bool empty() { return false; }
        inline static size_type max_size() { return N; }

        // direct access to data
        inline const T* data() const { return data_; }

        // assignment with possible type conversion
        template <typename T2>
        inline self_t& operator= (const tiny_vector<T2,N>& rhs) {
            detail::inline_< detail::assign< T,
                typename tiny_vector<T2,N>::const_iterator, iterator>, N>
            ::apply_seq(rhs.begin(), begin());
            return *this;
        }

        // default element-by-element comparisons
        inline bool operator==(const self_t& rhs) const {
            return detail::inline_< detail::equal<T, const_iterator>, N>
                ::compute_seq(begin(), rhs.begin());
        }
        inline bool operator<(const self_t& rhs) const {
            return detail::inline_< detail::less<T, const_iterator>, N>
                ::compute_seq(begin(), rhs.begin());
        }


        // -------------------------------------
        // specific operators for vectors
        // -------------------------------------

        // assignment with possible type conversion 
        // and degrade/upgrade
        template <typename T2, std::size_t N2>
        inline self_t& operator= (const tiny_vector<T2,N2>& rhs) {
            BOOST_STATIC_ASSERT(
                (((N + 1) == N2) || ((N - 1) == N2)) &&
                (N != N2) /* just in case :-) */ );
            
            detail::inline_< detail::assign< T,
                typename tiny_vector<T2,N2>::const_iterator, iterator>, 
                (N > N2) ? N2 : N >
            ::apply_seq(rhs.begin(), begin());
            
            return detail::change_grade< self_t, tiny_vector<T2,N2> >
                ::apply( *this, rhs );
        }

        // unary plus
        inline self_t operator +() const {
            // do nothing
            return *this;
        }
        // unary minus
        inline self_t operator -() const {
            self_t nrv;
            detail::inline_<
                detail::negate_assign<T, const_iterator, iterator>, N >
            ::apply_seq(begin(), nrv.begin());
            return nrv;
        }

        // this + vector
        inline self_t& operator += (const self_t& rhs) {
            detail::inline_< detail::add< T,
                const_iterator, iterator>, N>
            ::apply_seq(rhs.begin(), begin());
            return *this;
        }
        // vector operator+(vector, const vector&) automatically
        // generated by addable.

        // this - vector
        inline self_t& operator -= (const self_t& rhs) {
            detail::inline_< detail::sub< T,
                const_iterator, iterator>, N>
            ::apply_seq(rhs.begin(), begin());
            return *this;
        }
        // vector operator-(vector, const vector&) automatically
        // generated by subtractable.

        // this * scalar
        inline self_t& operator *= (const T& v) {
            detail::inline_<
                detail::mul_v<T, iterator>, N>
            ::apply_value(begin(), v);
            return *this;
        }
        // vector operator*(vector, const T&) and
        // vector operator*(const T&, vector) auto-generated
        // by multipliable.

        // this / scalar value
        inline self_t& operator /= (const T& v) {
            detail::inline_<
                detail::div_v<T, iterator>, N>
            ::apply_value(begin(), detail::div_v<T, iterator>::divisor(v));
            return *this;
        }
        // vector operator/(vector, const T&) auto-generated
        // by dividable.

        // dot product
        inline T operator * (const self_t& rhs) const {
            return detail::inline_< detail::dot_pr<T, const_iterator>, N>
                ::compute_seq(begin(), rhs.begin());
        }

        // Length squared of the vector = vec . vec
        inline T sq_length() const {
            return ((*this) * (*this));
        }

        // Length of the vector = sqrt( vec . vec )
        inline length_type length() const {
            return sqrt_t<T>::apply( sq_length() );
        }

        // Returns the unit vector in the direction of this vector.
        // Attempting to normalize a zero-vector will result in a divide by
        // zero error.  This is as it should be. Fix the calling code.
        inline self_t unit() const {
            return ((*this) / length());
        }

        // Normalize the vector so that it has length unity.
        // The same caution as above.
        inline void normalize() {
            (*this) /= length();
        }

        // stores cross product in this vector
        //  NB. generalized cross product calculation
        //  for odd vector grades would be nice
        inline self_t& operator ^= (const self_t& rhs)
        {
            //BOOST_STATIC_ASSERT( (N >= 3) && (N % 2) );
            BOOST_STATIC_ASSERT( N == 3 );
            
            T c0 = data_[1]*rhs[2] - data_[2]*rhs[1];
            T c1 = data_[2]*rhs[0] - data_[0]*rhs[2];
            T c2 = data_[0]*rhs[1] - data_[1]*rhs[0];
            
            data_[0] = c0;
            data_[1] = c1;
            data_[2] = c2;

            return (*this);
        }

        // named components of vector
#define TINY_VECTOR_COMPONENT( NAME, I )                    \
        inline reference NAME()                             \
            { BOOST_STATIC_ASSERT(N > I); return data_[I]; }\
        inline const_reference NAME() const                 \
            { BOOST_STATIC_ASSERT(N > I); return data_[I]; }

        TINY_VECTOR_COMPONENT( x, 0 );
        TINY_VECTOR_COMPONENT( y, 1 );
        TINY_VECTOR_COMPONENT( z, 2 );
        TINY_VECTOR_COMPONENT( w, 3 );
#undef TINY_VECTOR_COMPONENT
    private:
        // check range (may be private because it is static)
        static void rangecheck (size_type i) {
            if (i >= size()) { throw std::range_error("tiny_vector"); }
        }
    };


    // predefined (possibly) useful vector types
    typedef tiny_vector< float,  2 > vector2f;
    typedef tiny_vector< float,  3 > vector3f;
    typedef tiny_vector< float,  4 > vector4f;

    typedef tiny_vector< double, 2 > vector2d;
    typedef tiny_vector< double, 3 > vector3d;
    typedef tiny_vector< double, 4 > vector4d;
} // namespace math

#undef MAX_TINY_VECTOR_CTOR_PARAMS

#endif // _TINY_VECTOR_HPP_
